@@ -1,6 +1,6 @@
 /*
-   mod_ruid2 0.9.7
-   Copyright (C) 2009-2012 Monshouwer Internet Diensten
+   mod_ruid2 0.9.8
+   Copyright (C) 2009-2013 Monshouwer Internet Diensten
 
    Author: Kees Monshouwer
 
@@ -48,7 +48,7 @@
 #include <sys/capability.h>
 
 #define MODULE_NAME		"mod_ruid2"
-#define MODULE_VERSION		"0.9.7"
+#define MODULE_VERSION		"0.9.8"
 
 #define RUID_MIN_UID		100
 #define RUID_MIN_GID		100
@@ -102,8 +102,9 @@ module AP_MODULE_DECLARE_DATA ruid2_module;
 
 static int mode_stat_used	= RUID_MODE_STAT_NOT_USED;
 static int chroot_used		= RUID_CHROOT_NOT_USED;
+static int cap_mode		= RUID_CAP_MODE_KEEP;
 
-static int coredump, cap_mode, root_handle;
+static int coredump, root_handle;
 static const char *old_root;
 
 static gid_t startup_groups[RUID_MAXGROUPS];
@@ -293,7 +294,7 @@ static const command_rec ruid_cmds[] = {
 
 	AP_INIT_TAKE1 ("RMode", set_mode, NULL, RSRC_CONF | ACCESS_CONF, "Set mode to config or stat (default: config)"),
 	AP_INIT_TAKE2 ("RUidGid", set_uidgid, NULL, RSRC_CONF | ACCESS_CONF, "Minimal uid or gid file/dir, else set[ug]id to default (User,Group)"),
-	AP_INIT_ITERATE ("RGroups", set_groups, NULL, RSRC_CONF | ACCESS_CONF, "Set aditional groups"),
+	AP_INIT_ITERATE ("RGroups", set_groups, NULL, RSRC_CONF | ACCESS_CONF, "Set additional groups"),
 	AP_INIT_TAKE2 ("RDefaultUidGid", set_defuidgid, NULL, RSRC_CONF, "If uid or gid is < than RMinUidGid set[ug]id to this uid gid"),
 	AP_INIT_TAKE2 ("RMinUidGid", set_minuidgid, NULL, RSRC_CONF, "Minimal uid or gid file/dir, else set[ug]id to default (RDefaultUidGid)"),
 	AP_INIT_TAKE2 ("RDocumentChRoot", set_documentchroot, NULL, RSRC_CONF, "Set chroot directory and the document root inside"),
@@ -318,6 +319,12 @@ static int ruid_init (apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server
 		apr_pool_userdata_set((const void *)1, userdata_key, apr_pool_cleanup_null, s->process->pool);
 	} else {
 		ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, NULL, MODULE_NAME "/" MODULE_VERSION " enabled");
+
+		/* MaxRequestsPerChild MUST be 1 to enable drop capability mode */
+		if (ap_max_requests_per_child == 1) {
+			ap_log_error(APLOG_MARK, APLOG_NOTICE, 0, NULL, MODULE_NAME " is in drop capability mode");
+			cap_mode = RUID_CAP_MODE_DROP;
+		}
 	}
 
 	return OK;
@@ -352,16 +359,21 @@ static void ruid_child_init (apr_pool_t *p, server_rec *s)
 	}
 
 	/* setup chroot jailbreak */
-	if (chroot_used == RUID_CHROOT_USED) {
+	if (chroot_used == RUID_CHROOT_USED && cap_mode == RUID_CAP_MODE_KEEP) {
 		if ((root_handle = open("/.", O_RDONLY)) < 0) {
 			root_handle = UNSET;
 			ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "%s CRITICAL ERROR opening root file descriptor failed (%s)", MODULE_NAME, strerror(errno));
+		} else if (fcntl(root_handle, F_SETFD, FD_CLOEXEC) < 0) {
+			ap_log_error(APLOG_MARK, APLOG_ERR, 0, NULL, "%s CRITICAL ERROR unable to set close-on-exec flag on root file descriptor (%s)", MODULE_NAME, strerror(errno));
+			if (close(root_handle) < 0)
+				ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL, "%s CRITICAL ERROR closing root file descriptor (%d) failed", MODULE_NAME, root_handle);
+			root_handle = UNSET;
 		} else {
 			/* register cleanup function */
 			apr_pool_cleanup_register(p, (void*)((long)root_handle), ruid_child_exit, apr_pool_cleanup_null);
 		}
 	} else {
-		root_handle = UNSET;
+		root_handle = (chroot_used == RUID_CHROOT_USED ? NONE : UNSET);
 	}
 
 	/* init cap with all zeros */
@@ -382,9 +394,6 @@ static void ruid_child_init (apr_pool_t *p, server_rec *s)
 	}
 	cap_free(cap);
 
-	/* MaxRequestsPerChild MUST be 1 to enable drop capability mode */
-	cap_mode = (ap_max_requests_per_child == 1 ? RUID_CAP_MODE_DROP : RUID_CAP_MODE_KEEP);
-
 	/* check if process is dumpable */
 	coredump = prctl(PR_GET_DUMPABLE);
 }
@@ -461,7 +470,7 @@ static int ruid_set_perm (request_rec *r, const char *from_func)
 	int groupsnr;
 
 	cap_t cap;
-	cap_value_t capval[4];
+	cap_value_t capval[3];
 
 	cap=cap_get_proc();
 	capval[0]=CAP_SETUID;
@@ -533,14 +542,9 @@ static int ruid_set_perm (request_rec *r, const char *from_func)
 	capval[2]=CAP_DAC_READ_SEARCH;
 	cap_set_flag(cap,CAP_EFFECTIVE,3,capval,CAP_CLEAR);
 
-	if (cap_mode == RUID_CAP_MODE_DROP) {
-		/* clear capabilities from permitted set (permanent) */
-		capval[3]=CAP_SYS_CHROOT;
-		cap_set_flag(cap,CAP_PERMITTED,4,capval,CAP_CLEAR);
-	}
-
 	if (cap_set_proc(cap)!=0) {
 		ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL, "%s CRITICAL ERROR %s>%s:cap_set_proc failed after setuid", MODULE_NAME, from_func, __func__);
+		retval = HTTP_FORBIDDEN;
 	}
 	cap_free(cap);
 
@@ -617,7 +621,34 @@ static int ruid_setup (request_rec *r)
 /* run in map_to_storage hook */
 static int ruid_uiiii (request_rec *r)
 {
-		return ruid_set_perm(r, __func__);
+	if (!ap_is_initial_req(r)) {
+		return DECLINED;
+	}
+
+	int retval = ruid_set_perm(r, __func__);
+
+	int ncap;
+	cap_t cap;
+	cap_value_t capval[4];
+
+	/* clear capabilities from permitted set (permanent) */
+	if (cap_mode == RUID_CAP_MODE_DROP) {
+		cap=cap_get_proc();
+		capval[0]=CAP_SETUID;
+		capval[1]=CAP_SETGID;
+		capval[2]=CAP_DAC_READ_SEARCH;
+		ncap = 2;
+		if (root_handle == UNSET) capval[ncap++] = CAP_SYS_CHROOT;
+		cap_set_flag(cap,CAP_PERMITTED,ncap,capval,CAP_CLEAR);
+
+		if (cap_set_proc(cap)!=0) {
+			ap_log_error (APLOG_MARK, APLOG_ERR, 0, NULL, "%s CRITICAL ERROR %s:cap_set_proc failed after setuid", MODULE_NAME, __func__);
+			retval = HTTP_FORBIDDEN;
+		}
+		cap_free(cap);
+	}
+
+	return retval;
 }
 
 
